
#include <mpnt/mpnt.h>
using namespace MpNT;

#include "DES.h"

// A DES key consists of 64 binary digits ("0"s or "1"s) of which 56 bits are randomly generated and
// used directly by the algorithm. The other 8 bits, which are not used by the algorithm, may be used
// for error detection. The 8 error detecting bits are set to make the parity of each 8-bit byte of the key
// odd, i.e., there is an odd number of "1"s in each 8-bit byte1. One bit in each 8-bit byte of the KEY may
// be utilized for error detection in key generation, distribution and storage. Bits 8, 16,..., 64 are for
// use in assuring that each byte is of odd parity.
MpInt DES::generate_key() {
	MpInt key;
	key.random_bits(64);
	int s = 0;
	for (int i = 1; i<=64; i++) {
		if (i % 8 == 0) {
			key.set_bit(i-1, s % 2 == 0);
			s = 0;
		} else {
			s += (int)key.get_bit(i-1);
		}
	}
	return key;
}

bool DES::check_key(const MpInt& key) {
	int s = 0;
	for (int i = 1; i<=64; i++) {
		s += (int)key.get_bit(i-1);
		if (i % 8 == 0) {
			if (s % 2 == 0) {
				return false;
			}
			s = 0;
		}
	}
	return true;
}

MpInt DES::pc1(const MpInt& key) {
	MpInt result;
	for (int i = 0; i<56; i++) {
		result.set_bit(i, key.get_bit(PC1[i]-1));
	}
	return result;
}

MpInt DES::pc2(const MpInt& cidi) {
	MpInt result;
	for (int i = 0; i<48; i++) {
		result.set_bit(i, cidi.get_bit(PC2[i]-1));
	}
	return result;
}

MpInt DES::lsi(const MpInt& cd, int i) {
	if (i == 1 || i == 2 || i == 9 || i == 16) {
		MpInt result(cd);
		shl(result, result, 2);
		result.set_bit(0, result.get_bit(28));
		result.set_bit(1, result.get_bit(29));
		result.set_bit(28, result.get_bit(56));
		result.set_bit(29, result.get_bit(57));
		result.set_bit(56, false);
		result.set_bit(57, false);
		return result;
	} else {
		MpInt result(cd);
		shl(result, result, 1);
		result.set_bit(0, result.get_bit(28));
		result.set_bit(28, result.get_bit(56));
		result.set_bit(56, false);
		return result;
	}
}

void DES::compute_round_keys(const MpInt key, vector<MpInt> &sub_keys, int rounds) {
	sub_keys.clear();
	MpInt cd = DES::pc1(key);
	for (int i = 1; i<=16; i++) {
		cd = lsi(cd, i);
		sub_keys.push_back(pc2(cd));
	}
}

MpInt DES::p0(const MpInt& x) {
	MpInt result;
	for (int i = 0; i<64; i++) {
		result.set_bit(i, x.get_bit(P0[i]-1));
	}
	return result;
}

MpInt DES::p0_inv(const MpInt& x) {
	MpInt result;
	for (int i = 0; i<64; i++) {
		result.set_bit(P0[i]-1, x.get_bit(i));
	}
	return result;	
}

MpInt DES::rev(const MpInt& uv) {
	MpInt u; // leftmost 32 bits of cd
	MpInt v; // rightmost 32 bits of cd
	shr(u, uv, 32);
	MpInt mask;
	mask.set_bit(32, true);
	dec(mask, mask);
	bit_and(v, uv, mask);
	MpInt vu = v;
	shl(vu, vu, 32);
	bit_or(vu, vu, u);
	return vu;
}

MpInt DES::e(const MpInt& v) {
	MpInt result;
	for (int i = 0; i<48; i++) {
		result.set_bit(i, v.get_bit(E[i]-1));
	}
	return result;
}

MpInt DES::p(const MpInt& x) {
	MpInt result;
	for (int i = 0; i<32; i++) {
		result.set_bit(i, x.get_bit(P[i]-1));
	}
	return result;
}

MpInt DES::s(const MpInt& x) {
	MpInt result;
	for (int i = 0; i<8; i++) {
		bool wi[6];
		for (int j = 0; j<6; j++) {
			wi[j] = x.get_bit(6*(7-i) + j);
		}
		int row = wi[5]*2 + wi[0];
		int column = wi[4]*8 + wi[3]*4 + wi[2]*2 + wi[1];
		shl(result, result, 4);
		bit_or(result, result, MpInt(M[i][row][column]));
	}
	return result;
}

MpInt DES::f(const MpInt& v, const MpInt& ki) {
	MpInt temp;
	bit_xor(temp, e(v), ki);
	return p(s(temp));
}

MpInt DES::tki(const MpInt& uv, const MpInt& ki) {
	MpInt u; // leftmost 32 bits of cd
	MpInt v; // rightmost 32 bits of cd
	shr(u, uv, 32);
	MpInt mask;
	mask.set_bit(32, true);
	dec(mask, mask);
	bit_and(v, uv, mask);
	MpInt result = v;
	shl(result, result, 32);
	MpInt temp;
	bit_xor(temp, u, f(v, ki));
	bit_or(result, result, temp);
	return result;
}

MpInt DES::encrypt(const MpInt& x, const vector<MpInt>& round_keys) {
	MpInt y = p0(x);
	for (unsigned int i=0; i<round_keys.size(); i++) {
		y = tki(y, round_keys[i]);
	}
	return p0_inv(rev(y));
}

MpInt DES::decrypt(const MpInt& y, const vector<MpInt>& round_keys) {
	MpInt x = p0(y);
	for (int i=round_keys.size()-1; i>=0; i--) {
		x = tki(x, round_keys[i]);
	}
	return p0_inv(rev(x));
}
